from pydantic import BaseModel, DirectoryPath, constr, field_validator, model_validator, NewPath
from typing import Optional
import re
import yaml

class ConfigError(Exception):
    def __init__(self, msg):
        message_primer = "There appears to be a problem with the values entered:\n"
        self.message = message_primer + msg
        super().__init__(self.message)

class ConfigStrategy(BaseModel):
    processed_directory: DirectoryPath
    file_name_pattern: str
    file_content_pattern: Optional[str] = None
    terms: dict[str, str]
    export_format: str
    export_path: NewPath
    export_csv_divider: str = ';'
    # terms_patterns_group is created from 'terms', see @model_validator
    terms_patterns_group: dict[str, tuple[re.Pattern, int, int]] = None
    matchall_maxlength: int = 100

    @field_validator('file_name_pattern', 'export_format')
    @classmethod
    def non_empty_string(cls, v: str):
        assert v != '', 'Must be a non-empty string'
        return v.strip()
    
    @classmethod
    def compile_regex(cls, p: str) -> re.Pattern:
        try:
            rgx = re.compile(p)
            return rgx
        except:
            raise ConfigError(f"Concerning pattern {p}: this string cannot be used as a regex pattern!")

    @field_validator('file_name_pattern', 'file_content_pattern')
    @classmethod
    def selection_must_be_regex(cls, v: str):
        v = v.strip()
        cls.compile_regex(v) # if unsuccessful an error is thrown
        return v

    @model_validator(mode='after')
    def check_terms_and_patterns(self):
        # process terms_and_patterns 
        processed_tp = {term:self.process_terms(pattern) for term, pattern in self.terms.items()}
        self.terms_patterns_group = processed_tp
        return self
    
    @field_validator('export_format')
    @classmethod
    def validate_export_format(cls, ef: str):
        valid_formats = {"csv", "xlsx"}
        if ef not in valid_formats:
            raise ConfigError(msg=f"Concerning '{ef}': Export format must conform to one of these options: {valid_formats}!")
        return ef
    
    @classmethod
    def is_regex(cls, patternstring: str) -> bool:
        #breakpoint()
        try: 
            re.compile(patternstring)
        except:
            return False
        return True
    
    # process_terms
    # This function has the following jobs:
    #   - check if patternstrings can be converted to regex patterns (type re.Pattern)
    #   - check if patternstrings already contain a "matchall pattern" (.*), as these are not allowed 
    #   - create capture groups if divider is present; if present:
    #       - check if divider occurs more than once, as this is not allowed
    #       - replace the divider by a capture group matching all ("matchall pattern")
    #       - return the index of the (one and only) capture group representing the matchall pattern
    #       - return the total number of capture groups
    # Return value:
    #   The function returns a tuple of (re.Pattern, int, int) containing the compiled pattern,
    #   the index of the group containing the (one and only) matchall pattern, and
    #   the number of capture groups present.
    #   In case no capture groups have been formed, the second and third integers are set to -1.
    def process_terms(cls, patternstring: str, divider: str = "@@@") -> tuple[re.Pattern, int, int]:
        # helper to check if pattern only consists of a matchall pattern
        def matchall_only(s) -> bool:
            return re.search("\.\*", s) and len(s) == 2
        # check if matchall pattern is present (as this is not allowed)
        if matchall_only(patternstring):
            raise ConfigError(msg=f"The string '{patternstring}' only contains the matchall-pattern '.*' and can therefore not be processed.")
        # divider_hits counts the number of divider in the string; only one is allowed (see below)
        divider_hits = len(re.findall(divider, patternstring))
        # check the number of occurrences of divider
        if divider_hits > 1:
            # as this is not implemented, throw an error
            raise ConfigError(msg=f"Each term must correspond to either *one regex pattern* or *two regex patterns divided by '{divider}'*!")
        if divider_hits == 0:
            # return without capture groups
            return (cls.compile_regex(patternstring), -1, -1)
        # process the patternstrings divided by divider
        multiple_patternstrings = re.split(divider, patternstring)
        # do any of the patternstrings only contain a matchall pattern?
        if any([matchall_only(p) for p in multiple_patternstrings]):
            raise ConfigError(msg=f"At least one of {multiple_patternstrings!r} only consists of a matchall-pattern '.*' and can therefore not be processed.")
        # is any of the patternstrings of length 0?
        lenx = [len(i) for i in multiple_patternstrings]
        lenx0 = [l == 0 for l in lenx]
        # if yes
        if any(lenx0):
            # the first?
            if lenx0[0]:
                return (cls.compile_regex(f"(.*)({multiple_patternstrings[1]})"), 1, 2)
            # the second?
            return (cls.compile_regex(f"({multiple_patternstrings[0]})(.*)"), 2, 2)
        # none of the patternstrings empty? return three groups
        return (cls.compile_regex(f"({multiple_patternstrings[0]})(.*)({multiple_patternstrings[1]})"), 2, 3)

class Config(BaseModel):
    title: str
    strategies: dict[str, ConfigStrategy]
    


__version__ = '0.0.2'
import yaml
import re
import os
from .pydantic_models import Config, ConfigStrategy, ConfigError
from .database import Base, DbRun, DbStrategy, DbProcessedFile, DbMatch
from .utils import file_sha256
from .processors import DocumentProcessor, PDFProcessor
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session
import argparse

def read_config_file(config_file: str) -> Config:
    try:
        with open(config_file, "r") as f:
            cfg_raw = yaml.safe_load(f)
        cfg = Config(**cfg_raw)
        return cfg
    except Exception as e:
        print(f"{e}")
        
def find_documents(directory, file_name_pattern) -> list[str]:
    pattern = re.compile(file_name_pattern)
    return [os.path.join(directory, f) for f in os.listdir(directory) if pattern.match(f)]

def process_strategy(strategy_name: str, cs: ConfigStrategy, session: Session, run_id: int, 
                     processor: DocumentProcessor):
    # save strategy info
    d_strategy = DbStrategy(run_id = run_id,
                            name = strategy_name,
                            processed_directory = str(cs.processed_directory),
                            file_name_pattern = cs.file_name_pattern,
                            file_content_pattern = cs.file_content_pattern,
                            export_format = cs.export_format,
                            export_path = str(cs.export_path),
                            export_csv_divider = cs.export_csv_divider)
    session.add(d_strategy)
    session.commit()
    
    # loop thru files
    documents = find_documents(cs.processed_directory, cs.file_name_pattern)
    if len(documents) < 1:
        raise ConfigError(f"No matching documents found with pattern {cs.file_name_pattern} in directory {cs.processed_directory}!")
    
    for doc in documents:
        pr = processor(doc, cs.matchall_maxlength)
        # if file_content_pattern is given and if that pattern is not found in the document, skip the document
        if cs.file_content_pattern:
            if not pr.contains(cs.file_content_pattern):
                # skip to next document without any further ado
                continue
        # save file info
        d_file = DbProcessedFile(strategy_id = d_strategy.id,
                                 filename = doc,
                                 sha256 = file_sha256(doc))
        session.add(d_file)
        session.commit()
        
        # get content from file
        terms_content = pr.terms_content(cs.terms_patterns_group)
        
        # save content into database
        for term, content in terms_content.items():
            dm = DbMatch(file_id = d_file.id, term = term, content = content)
            session.add(dm)
            session.commit()
    
def main(config_file: str, db_file: str):
    config = read_config_file(config_file)
    engine = create_engine(f'sqlite:///{db_file}')
    Base.metadata.create_all(engine)
    
    Session = sessionmaker(bind=engine)
    session = Session()
    
    # save run info
    run = DbRun(title = config.title,
              yml_filename = config_file,
              yml_sha256 = file_sha256(config_file))
    session.add(run)
    session.commit()
    # pass session on
    
    for strategy_name, strategy in config.strategies.items():
        process_strategy(strategy_name, strategy, session, run.id, PDFProcessor)

    session.close()
    
def cli():
    parser = argparse.ArgumentParser(description='Process PDFs according to a YAML configuration.')
    parser.add_argument('-c', '--config', default='config.yml', help='Path to configuration YAML file')
    parser.add_argument('-d', '--database', default='matches.db', help='Path to SQLite database file')
    args = parser.parse_args()
    
    # check if config exists!
    if not os.path.isfile(args.config):
        raise Exception("No configuration file provided! Please add one using -c!")

    main(args.config, args.database)
    
if __name__ == "__main__":
    cli()
    
    


import fitz
import re

class DocumentProcessor:
    text: str
    matchall_maxlength: int
    result: dict[str, str]
    
    def __init__(self, file_path, matchall_maxlength):
        self.extract_text(file_path = file_path)
        self.matchall_maxlength = matchall_maxlength
    
    def extract_text(self, file_path):
        raise NotImplementedError("This method should be implemented by subclasses.")
    
    # function to process the "terms and pattern group" from ConfigStrategy
    def terms_patterns(self, tap: dict[str, tuple[re.Pattern, int, int]]):
        result = {}
        for term, pattern_tpl in tap.items():
            p = pattern_tpl[0]
            mo = re.search(p, self.text)
            content = None
            has_groups = pattern_tpl[1] > 0
            # have patterns been found at all?
            if mo:
                # are groups present?
                if has_groups:
                    matchall_index = pattern_tpl[1]
                    number_of_groups = pattern_tpl[2] # also p.groups
                    content = mo.group(matchall_index)
                    # in case only two groups present: limit length of matched text
                    if len(content) > self.matchall_maxlength and number_of_groups == 2:
                        if matchall_index == 1:
                            content = content[-self.matchall_maxlength:]
                        else:
                            content = content[:self.matchall_maxlength]
                # no groups
                else:
                    # mos: indices of matched text
                    mos = mo.span()
                    content = self.text[mos[0]:mos[1]]
            
            # document result
            result[term] = content
        # put into class attribute
        self.result = result
        
    def terms_content(self, tap) -> dict[str, str]:
        self.terms_patterns(tap)
        return self.result
                
    def contains(self, patternstring: str) -> bool:
        pattern = re.compile(patternstring)
        if pattern.search(self.text):
            return True
        else:
            return False
        
class PDFProcessor(DocumentProcessor):
    def extract_text(self, file_path):
        text = ""
        with fitz.open(file_path) as doc:
            for page in doc:
                text += page.get_text()
        self.text = text

# Placeholder for future extensions
# class MarkdownProcessor(DocumentProcessor):
#     ...
